unit Unsigned;

{$Q-}

interface

  { Unsigned 32-bit integers.
    Delphi Pascal does not support unsigned 32-bit integers, but
    only signed 32-bit integers. The difference matters when comparing
    values, doing division, converting to and from a string, or
    checking for overflow errors. For the latter, the simplest
    solution is to turn of overflow checking with the $Q- compiler
    directive. This unit supplies the other functions.

    The names are a bit terse, I admit, but UnsignedLessOrEqual
    is too long, and I could not think of anything in between.

    Some of code is a little hard to read because there are three
    distinct sets of source code intertwingled.

    1. 16-bit Win 3.1 code.
    2. 32-bit Win 3.1 code. Some routines are written in assembly
       language and take advantage of 32-bit instructions when
       running on a 386 class CPU or better.
    3. 32-bit Win 32 code. The Delphi 2.0 compiler passes arguments
       in registers, which allows for much faster and simpler code.

    Ray Lischner
    Tempest Software
    5 January 1996
  }

  { To help the human reader, declare 32-bit unsigned integer values
    with the TUnsigned type. It is just a LongInt, but it notifies
    the reader of the programmer's intent. }
  type
    TUnsigned = LongInt;

  { Comparison of unsigned 32-bit integers. }
  function ULT(A, B: TUnsigned): Boolean;
  function ULE(A, B: TUnsigned): Boolean;
  function UGT(A, B: TUnsigned): Boolean;
  function UGE(A, B: TUnsigned): Boolean;

  { Result < 0 if A<B
    Result = 0 if A=B
    Result > 0 if A>B
    Do not make any other assumptions about the Result. }
  function CompareUnsigned(A, B: TUnsigned): Integer;

  { Convert between a string and an unsigned long integer. }
  function StrToUInt(const Str: string): TUnsigned;
  function UIntToStr(UInt: TUnsigned): string;

  { Unsigned, 32-bit division & remainder. }
  function UDiv(Dividend, Divisor: TUnsigned): TUnsigned;
  function UMod(Dividend, Divisor: TUnsigned): TUnsigned;

  { You can perform 32-bit unsigned arithmetic using Delphi's
    Comp type. This is a 64-bit signed integer type, but it
    uses floating point math, so it is slow. In cases where
    development time is more important, however, it can be a
    valuable aid. The UIntToComp and CompToUint functions
    convert between Comp and TUnsigned, performing overflow
    checking when converting from Comp to TUnsigned. }
  function UIntToComp(UInt: TUnsigned): Comp;
  function CompToUInt(Value: Comp): TUnsigned;

implementation

uses
  {$ifndef WIN32}FltMath,{$endif}
  SysUtils;

  { Because the Delphi 2.0 compiler passes arguments in registers,
    and 32-bit instructions are available, it is best to code
    the Win32 comparison routines explicitly. For Delphi 1.0 and
    Windows 3.1, call the common CompareUnsigned routine, for simplicity. }
  function ULT(A, B: TUnsigned): Boolean;
{$ifdef WIN32}
  assembler;
  asm
    mov  ecx, eax
    xor  eax, eax    { assume A >= B by setting Result := False}
    cmp  ecx, edx    { compare A and B }
    jnb  @@1         { if A >= B then return False }
    inc  eax         { if A < B then return True }
  @@1:
  end;
{$else}
  begin
    Result := CompareUnsigned(A, B) < 0;
  end;
{$endif}

  function ULE(A, B: TUnsigned): Boolean;
{$ifdef WIN32}
  assembler;
  asm
    mov  ecx, eax
    xor  eax, eax    { assume A > B by setting Result := False}
    cmp  ecx, edx    { compare A and B }
    jnbe @@1         { if A > B then return False }
    inc  eax         { if A <= B then return True }
  @@1:
  end;
{$else}
  begin
    Result := CompareUnsigned(A, B) <= 0;
  end;
{$endif}

  function UGT(A, B: TUnsigned): Boolean;
{$ifdef WIN32}
  assembler;
  asm
    mov  ecx, eax
    xor  eax, eax    { assume A <= B by setting Result := False}
    cmp  ecx, edx    { compare A and B }
    jna  @@1         { if A <= B then return False }
    inc  eax         { if A > B then return True }
  @@1:
  end;
{$else}
  begin
    Result := CompareUnsigned(A, B) > 0;
  end;
{$endif}


  function UGE(A, B: TUnsigned): Boolean;
{$ifdef WIN32}
  assembler;
  asm
    mov  ecx, eax
    xor  eax, eax    { assume A < B by setting Result := False}
    cmp  ecx, edx    { compare A and B }
    jnae @@1         { if A < B then return False }
    inc  eax         { if A >= B then return True }
  @@1:
  end;
{$else}
  begin
    Result := CompareUnsigned(A, B) >= 0;
  end;
{$endif}

  { Compare two unsigned 32-bit integers.  Return <0, =0, or >0.
    Start by comparing the high-order words.  If they are not equal,
    then we are done and can return -1 or +1.  If the high order words
    are equal, then we need to compare the low-order words.  If they
    are not equal, then return -1 or +1.  If the low order words are
    also equal, then return 0.

    If anyone has a faster way to do this, I am open to suggestions.
    Of course, using 386 instructions would certainly be faster and
    smaller, but also not functional on 286 machines.

    Just for reference, here is a 32-bit implementation:

  function CompareUnsigned32(A, B: TUnsigned): Integer; assembler;
  asm
    xor ax, ax
    db 66h
    mov cx, A.Word[0]        ; mov ecx, A.DWord[0]
    db 66h
    mov dx, B.Word[0]        ; mov edx, B.DWord[0]
    db 66h
    sub cx, dx               ; sub ecx, edx
    ja @Greater
    je @End
    dec ax
    jmp @End
  @Greater:
    inc ax
  @End:
  end;
}

  function CompareUnsigned(A, B: TUnsigned): Integer; assembler;
  asm
{$ifdef WIN32}
    mov ecx, eax              { Copy A(EAX) to ECX }
    xor eax, eax              { Assume equality }
    sub ecx, edx              { Compare A and B }
    je @End                   { A = B; so return 0 }
    ja @Greater
  @Less:
    dec eax                   { A < B; so return -1 }
    jmp @End
{$else}
    cmp Test8086, 3
    jl  @Slow
    xor ax, ax                { 80386 implementation }
    db 66h
    mov cx, A.Word[0]         { mov ecx, A.DWord[0] }
    db 66h
    mov dx, B.Word[0]         { mov edx, B.DWord[0] }
    db 66h
    sub cx, dx                { sub ecx, edx }
    ja @Greater
    je @End
    dec ax
    jmp @End
  @Slow:                      { slow 8086 implementation, with 16-bit integers }
    xor ax, ax                { Start by assuming equality }
    mov cx, A.Word[2]         { First compare the high order words }
    mov dx, B.Word[2]
    sub cx, dx
    ja @Greater               { A.Hi > B.Hi }
    jb @Less                  { A.Lo < B.Lo }
    { Otherwise, A.Hi = B.Hi, so compare the low-order words }
    mov cx, A.Word[0]
    mov dx, B.Word[0]
    sub cx, dx
    je @End                   { A = B; so return 0 }
    ja @Greater
  @Less:
    dec ax                    { A < B; so return -1 }
    jmp @End
{$endif}
  @Greater:
    inc ax                    { A > B; so return +1 }
  @End:
  end;

  { Convert a string to an unsigned long integer, just like StrToInt,
    but it does not recognize a leading sign, and it can produce any
    32-bit, unsigned value. }
  function StrToUInt(const Str: string): TUnsigned;
  var
    I: Integer;
    Val: 0..9;
  label
    Error;
  begin
    { StrToInt can already handle hexadecimal unsigned values. }
    if Str[1] = '$' then
    begin
      Result := StrToInt(Str);
      Exit;
    end;

    Result := 0;
    for I := 1 to Length(Str) do
      if Str[I] in ['0'..'9'] then
      begin
        Val := Ord(Str[I]) - Ord('0');
        Result := Result * 10;
        { The default overflow checking checks too much.
          Instead, all that is important is to know whether the
          carry bit has been set, which indicates an unsigned overflow. }
        asm
          jc Error
        end;
        Result := Result + Val;
        asm
          jc Error
        end;
      end
      else
      begin
      Error:
        raise EConvertError.CreateFmt('''%s'' is not a valid unsigned integer value', [Str]);
      end;
  end;

  { Reverse a string, that is, convert 'Foobar' to 'rabooF'. }
  procedure ReverseStr(var Str: string);
  var
    I, J: Integer;
    Tmp: Char;
  begin
    I := 1;
    J := Length(Str);
    while I < J do
    begin
      Tmp := Str[J];
      Str[J] := Str[I];
      Str[I] := Tmp;
      Inc(I);
      Dec(J);
    end;
  end;

  { Convert a 32-bit unsigned value to a string. It is simplest to
    generate the digits in reverse order and then reverse the
    entire string to get the final result. }
  function UIntToStr(UInt: TUnsigned): string;
  var
    Digit: '0'..'9';
  begin
    if UInt = 0 then
      Result := '0'
    else
      Result := '';
    while UInt <> 0 do
    begin
      Digit := Chr(UMod(UInt, 10) + Ord('0'));
      AppendStr(Result, Digit);
      UInt := UDiv(UInt, 10);
    end;
    ReverseStr(Result);
  end;

  { To help convert Comp to TUnsigned, use a variant record that
    can hold both. The TUnsigned is the lower 32-bits of the Comp. }
  type
    TCompRecord = record
      case Boolean of
      True:  (VInt: array[0..1] of TUnsigned);
      False: (VComp: Comp);
    end;

  { Convert an unsigned 32-bit integer to a Comp. The Comp type
    is a 64-bit signed integer, and can be used for 32-bit
    unsigned arithmetic. }
  function UIntToComp(UInt: TUnsigned): Comp;
  var
    CompVar: TCompRecord;
  begin
    CompVar.VInt[0] := Uint;
    CompVar.VInt[1] := 0;
    Result := CompVar.VComp;
  end;

  { Convert a 64-bit signed integer to a 32-bit unsigned integer. }
  function CompToUInt(Value: Comp): TUnsigned;
  var
    CompVar: TCompRecord;
  begin
    CompVar.VCOmp := Value;
    if CompVar.VInt[1] <> 0 then
      raise EConvertError.Create('Unsigned integer overflow');
    Result := CompVar.VInt[0];
  end;

  { Doing 32-bit division on a 16-bit machine is a nightmare.
    Instead, convert to Comp and use floating point arithmetic
    to perform the division. Since Comp is 64 bits, the full
    precision is preserved. When running on a 32-bit machine,
    then use 32-bit integer arithmetic. }

  { Divide two 32-bit unsigned numbers to produce a 32-bit
    unsigned result. }
  function UDiv(Dividend, Divisor: TUnsigned): TUnsigned;
{$ifdef WIN32}
  assembler;
  asm
    mov ecx, edx
    xor edx, edx
    div ecx
  end;
{$else}
  var
    CDividend: Comp;
    CDivisor: Comp;
    OldRoundMode: TRoundMode;
  begin
    if Test8086 > 2 then
    begin
      asm
        db 66h
        mov ax, Dividend.Word[0]
        db 66h
        xor dx, dx
        db 66h
        mov cx, Divisor.Word[0]
        db 66h
        div cx
        { EAX holds the 32-bit result, which must be
          stuffed into DX:AX. }
        db 66h          { shld edx,eax,16 }
        dw 0a40fh       { But shld is a 386 instruction, so use the }
        dw 10c2h        { actual instruction codes instead. }
      end;
      Exit;
    end;

    { Explicitly check for zero to raise the integer divide by
      zero exception instead of letting the Comp division raise
      a floating point divide by zero exception. }
    if Divisor = 0 then
      raise EDivByZero.Create('Divide by zero in UDiv');
    CDividend := UIntToComp(Dividend);
    CDivisor  := UIntToComp(Divisor);
    OldRoundMode := SetRoundMode(rmTruncate);
    try
      Result := CompToUInt(CDividend / CDivisor);
    finally
      SetRoundMode(OldRoundMode);
    end;
  end;
  {$endif}

  { Divide two 32-bit unsigned numbers to produce a 32-bit
    unsigned remainder. }
  function UMod(Dividend, Divisor: TUnsigned): TUnsigned;
{$ifdef WIN32}
  assembler;
  asm
    mov ecx, edx
    xor edx, edx
    div ecx
    mov eax, edx
  end;
{$else}
  var
    Quotient: Comp;
    CDividend: Comp;
    CDivisor: Comp;
    OldRoundMode: TRoundMode;
  begin
    if Test8086 > 2 then
    begin
      asm
        db 66h
        mov ax, Dividend.Word[0]
        db 66h
        xor dx, dx
        db 66h
        mov cx, Divisor.Word[0]
        db 66h
        div cx
        { EDX holds the 32-bit result, which must be
          stuffed into DX:AX. }
        db 66h          { shld eax,edx,16 }
        dw 0a40fh       { But shld is a 386 instruction, so use the }
        dw 10d0h        { actual instruction codes instead. }
        db 66h
        shr dx, 16      { shr edx, 16 }
      end;
      Exit;
    end;

    { Explicitly check for zero to raise the integer divide by
      zero exception instead of letting the Comp division raise
      a floating point divide by zero exception. }
    if Divisor = 0 then
      raise EDivByZero.Create('Divide by zero in UMod');
    CDividend := UIntToComp(Dividend);
    CDivisor  := UIntToComp(Divisor);
    OldRoundMode := SetRoundMode(rmTruncate);
    try
      Quotient := CDividend / CDivisor;
    finally
      SetRoundMode(OldRoundMode);
    end;
    Result := CompToUInt(CDividend - (Quotient * CDivisor));
  end;
{$endif}

end.
{--=>revision-history<=--}
{1 UNSIGNED.PAS 3-Sep-95,16:25:16,`lisch' CHANGE=INITIAL                      }
{     Initial version.                                                        }
{2 UNSIGNED.PAS 5-Jan-96,14:55:28,`lisch' CHANGE=FULL_IMPLEMENTATION          }
{     Added conversion to strings, division, remainder.                       }
{3 UNSIGNED.PAS 5-Jan-96,21:59:58,`lisch' CHANGE=WIN32                        }
{     Completed support for WIN32 (Delphi 2.0), and improved 32-bit support fo}
{--=>revision-history<=--}
